#!/bin/bash
#
# kubeadm-upgrade version 0.0.1
#
# Copyright © 2019 Oracle Corp., Inc.  All rights reserved.
# Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl
#

function exitHandler() {
    exit_status=$?
    if [[ $exit_status -ne 0 ]]; then
        echo "-----------------------------------------------------" >> ${LOG_FILE}
        echo "Gracefully Ending After Failure" | tee -a ${LOG_FILE}
        echo "  - Time of failure: $(date)" | tee -a ${LOG_FILE}
        echo "  - Gathering Yum kubeadm, kubectl, kubelet Versions"
        echo "Yum Deps --------------------------------------------" >> ${LOG_FILE}
        yum list kubeadm >> ${LOG_FILE}
        yum list kubectl >> ${LOG_FILE}
        yum list kubelet >> ${LOG_FILE}
        echo "  - Gathering Docker Containers"
        echo "Docker Containers -----------------------------------" >> ${LOG_FILE}
        docker ps -a >> ${LOG_FILE}
        echo "  - Gathering Kublet Logs"
        echo "Kubelet Logs ----------------------------------------" >> ${LOG_FILE}
        journalctl -u kubelet.service --since "1 hour ago" >> ${LOG_FILE}
        echo "  - Gathering Docker Logs"
        echo "Docker Logs -----------------------------------------" >> ${LOG_FILE}
        journalctl -u docker.service --since "1 hour ago" >>  ${LOG_FILE}

        echo "[ERROR] Upgrade failed. See log file for more details: ${LOG_FILE}"
    fi
}

trap '' 2
trap exitHandler EXIT

: '
    Exit script with message and error code
    Parameter:
        $1: Exit message
        $2: Exit code
'
function do_exit() {
    stdout_and_log_to_file "$1"
    echo "Logs are available in file ${LOG_FILE}"
    exit $2
};
: '
    Output the message to stdout and append to log file
'
function stdout_and_log_to_file() {
    echo "$1"
    echo "[$(date)] $1" >> ${LOG_FILE}
};
: '
    Check if the system has enough space for an upgrade
    Parameter:
        input1: target directory
        inpout2: target size
'
function check_available_space() {
    src=$1
    # cut the first line(header). Getting available space in GB
    avail=$(df -BG --output=avail $src | awk 'FNR == 2' | grep -o -E '[0-9]+')
    stdout_and_log_to_file "Space available on the mount point $1: ${avail}GB"
    if [[ "${avail}" -lt $2 ]]; then
        do_exit "System does not have enough space for k8s cluster. Min should be at least ${2}GB"
    fi
};
: '
    check the exit code and exit with proper message if code is not zero
'
function check_exit_code() {
    if [ $1 -ne 0 ]; then
        do_exit "$2 failed" 1
    fi
};
: '
    print the usage
'
function print_usage() {
    echo "[NOTE] This script is used to upgrade kubernetes cluster from 1.9 version to 1.12 version." >&2
    echo "       This script is commonly used for both master and workers node" >&2
    echo "       and to backup or restore the master node in a cluster, is also available." >&2
    echo "usage:" >&2
    echo "  $0 upgrade" >&2
    echo "  $0 upgrade [--flannel | --dashboard]" >&2
    echo "  $0 backup directory" >&2
    echo "  $0 restore backup-file.tar" >&2
    exit 1

};

: '
    Check if a package in present in any of the yum repos
'
function check_package_in_yum() {
    local package=$1
    timeout ${TIMEOUT} yum install --downloadonly --downloaddir=/tmp/ ${package} > /dev/null
    if [ ! -f "/tmp/${package}.x86_64.rpm" ]; then
        do_exit "[ERROR] Please make sure that access to yum server is available for package upgrade" 1
    fi
    rm -rf /tmp/${package}.x86_64.rpm
};
: '
    Check whether a package is installed or  not
'
function isinstalled() {
    if yum list installed "$@" >/dev/null 2>&1; then
        true
    else
        false
    fi
};


: '
    Get the installed version of a package
    parameter:
        input: package name
        output: version format x.x
'
function get_package_version()
{
    version=$(rpm -q $1 | awk -F'-' '{print $2}' | awk -F'.' '{b=$1"."$2;print b}')
    echo $version
};

: '
    If a yum repo is enabled
    # remobve once we add rpms to addons
'
function repo_enabled() {
    state=$(yum repoinfo $1 | grep Repo-status |awk -F ':' '{print $2}'| awk '{$1=$1};1')
    if [ "$state" = "enabled" ]; then
        true
    else
        false
    fi
};

: '
    Check if the repo has current K8s version image
'
function check_repo {
    local repo="${1:-${REGISTRY}}"
    local image="${2:-pause-amd64:3.1}"

    local result=$(echo ""|docker pull ${repo}/${image} 2>&1|grep "login\|unauthorized: authentication required")
    if [ -n "${result}" ]; then
        stdout_and_log_to_file "[ERROR] Please login with valid credential to the ${repo}"
        stdout_and_log_to_file "        # docker login ${repo}"
        exit 1
    else
        if ! docker pull ${repo}/${image}; then
            stdout_and_log_to_file "[ERROR] docker cannot pull ${image} from ${repo} registry"
            exit 1
        fi
    fi
};

: '
    Check the compatibility of kubelet and kubectl rpm
'
function check_v1_11 {
    stdout_and_log_to_file "Checking kubelet and kubectl RPM ..."
    local kvers=$(rpm -qa kubeadm | awk -F- '{print $2 "-" $3}')
    for i in kubelet kubectl; do
        local version=$(rpm -qa | awk -v rpm=${i} -F- '$1 ~ rpm {print $2 "-" $3}')
        if [ "${version}" != "${kvers}" ]; then
            stdout_and_log_to_file "[INFO] yum install -y ${i}-${kvers}"
            yum install -y ${i}-$(echo ${kvers} | sed 's/\.x86_64//g')
        fi
    done

    if [ -f "${KUBELET_EXTRA}" ]; then
        rm -rf ${KUBELET_EXTRA}
        systemctl daemon-reload
    fi
};

: '
     update conf  file based on extra parameter for version 1.12
'
function registry_location12 {
    local registry="${1:-${REGISTRY}}"
    if [ -f "${KUBELET_EXTRA_NEW}" ]; then
        sed -i "s|KUBELET_EXTRA_ARGS=.*|KUBELET_EXTRA_ARGS=\"--fail-swap-on=false --pod-infra-container-image=${registry}/pause:3.1\"|g" ${KUBELET_EXTRA_NEW}
    fi
    systemctl daemon-reload
    systemctl restart kubelet
};

: '
     update conf  file based on extra parameter for version 1.10
'
function registry_location10 {
    local registry="${1:-}"
    if [ -n "${registry}" ]; then
        cat <<EOF >${KUBELET_EXTRA}
[Service]
Environment="KUBELET_EXTRA_ARGS=--pod-infra-container-image=${registry}/pause-amd64:3.1"
EOF
        chmod 644 ${KUBELET_EXTRA}
    else
        rm -f ${KUBELET_EXTRA}
    fi
    systemctl daemon-reload
    systemctl restart kubelet
};

: '
    update conf  file based on extra parameter for version 1.11
'
function registry_location11 {
    local registry="${1:-}"
    if [ -n "${registry}" ]; then
        if [ -f "${KUBELET_EXTRA_NEW}" ]; then
            sed -i "s|KUBELET_EXTRA_ARGS=.*|KUBELET_EXTRA_ARGS=\"--fail-swap-on=false --pod-infra-container-image=${registry}/pause-amd64:3.1\"|g" ${KUBELET_EXTRA_NEW}
        else
            cat <<EOF >${KUBELET_EXTRA}
[Service]
Environment="KUBELET_EXTRA_ARGS=--pod-infra-container-image=${registry}/pause-amd64:3.1"
EOF
            chmod 644 ${KUBELET_EXTRA}
        fi
    else
        rm -f ${KUBELET_EXTRA}
        if [ -f "${KUBELET_EXTRA_NEW}" ]; then
            sed -i "s|KUBELET_EXTRA_ARGS=.*|KUBELET_EXTRA_ARGS=\"--fail-swap-on=false\"|g" ${KUBELET_EXTRA_NEW}
        fi
  fi
  systemctl daemon-reload
  systemctl restart kubelet
};

: '
    Upgrade kubelet to 1.12 version
'
function upgrade_kubelet12 {
    local registry=${1:-${REGISTRY}}

    rm -rf ${KUBELET_EXTRA}
    stdout_and_log_to_file "Upgrading kubelet and kubectl now ... "
    check_v1_11

    if [ ! -f "/var/lib/kubelet/kubeadm-flags.env" ]; then
        cat <<EOF >/var/lib/kubelet/kubeadm-flags.env
KUBELET_KUBEADM_ARGS=--cgroup-driver=cgroupfs --cni-bin-dir=/opt/cni/bin --cni-conf-dir=/etc/cni/net.d --network-plugin=cni
EOF
        chmod 644 /var/lib/kubelet/kubeadm-flags.env
    fi

    registry_location12 "${registry}"
};

: '
    Upgrade Main
'
function upgrade_addons {
    local addons=${1:-}

    if ! is_master; then
	echo "[ERROR] upgrade addons command can only be run on master node"
	exit 1
    fi

    current_version=$(get_package_version "kubeadm")
    if [ "${current_version}" != "1.12" ]; then
	echo "[ERROR] Please first upgrade cluster by running 'kubeadm-upgrade.sh upgrade' before addons upgrade"
	exit 1
    fi

    if [ "${addons}" = "--flannel" ] || [ "${addons}" = "flannel" ]; then
        echo "Upgrading flannel"
        upgrade_flannel
    elif [ "${addons}" = "--dashboard" ] || [ "${addons}" = "dashboard" ]; then
        echo "Upgrading dashboard"
        upgrade_dashboard
    else
        echo "[ERROR] unrecognize options [${1}] for upgrade"
        echo ""
        print_usage
    fi
}

: '
    Upgrading flannel
'
function upgrade_flannel {
    local registry="${KUBE_REPO_PREFIX:-${REGISTRY}}"

    # flannel image pre-pull
    if [[ "${registry}" = "gcr.io/google_containers" ]] || [[ "${registry}" = "k8s.gcr.io" ]]; then
    	check_repo "quay.io/coreos" "flannel:v0.10.0"
    else
    	check_repo "${registry}" "flannel:v0.10.0"
    fi

    local flannel_pods=$(${KUBECTL} get pod | awk '/flannel/ {print $1}')
    echo $flannel_pods
    # delete the pod using old config file
    ${KUBECTL} delete -f /usr/local/share/kubeadm/flannel-ol-old.yaml --cascade=false
    for i in $flannel_pods; do
    	${KUBECTL} delete pod ${i} --cascade=false
    done
    ${KUBECTL} get pod
    local flannel="/usr/local/share/kubeadm/flannel-ol.yaml"

    if [ "${registry}" !=  "${REGISTRY}" ]; then
    	if [[ "${registry}" = "gcr.io/google_containers" ]] || [[ "${registry}" = "k8s.gcr.io" ]]; then
        	registry="quay.io/coreos"
        fi
    	cat ${flannel} | sed "s|${REGISTRY}|${registry}|g" > /tmp/flannel.yaml
    	flannel="/tmp/flannel.yaml"
    fi
    flannel_pods=$(grep "\-\-cluster-cidr" /etc/kubernetes/manifests/kube-controller-manager.yaml | awk -F= '{print $2}')
    sed -i "s|\([0-9]\{1,3\}\.\)\{3\}[0-9]\{1,3\}\/[0-9]*|${flannel_pods}|g" ${flannel}
    ${KUBECTL} create -f ${flannel}
}

: '
    Upgrading dashboard
'
function upgrade_dashboard {
    local registry="${KUBE_REPO_PREFIX:-${REGISTRY}}"

    # delete the old dashboard
    ${KUBECTL} delete -f /usr/local/share/kubeadm/kubernetes-dashboard-ol.yaml
    ${KUBECTL} delete secrets kubernetes-dashboard-certs

    export K8S_VERSION="${K8S_V12_IMAGE}"
    install_dashboard "${registry}"
}

: '
    Installing dashboard during upgrade
'
function install_dashboard {
    local registry="${1:-}"

    stdout_and_log_to_file ""
    stdout_and_log_to_file "Installing kubernetes-dashboard ..."
    stdout_and_log_to_file ""
    local dashboard=""
    if [ "${K8S_VERSION}" = "${K8S_V12_IMAGE}" ]; then
        dashboard="/usr/local/share/kubeadm/kubernetes-dashboard-self-certs.yaml"
    else
        dashboard="/usr/local/share/kubeadm/kubernetes-dashboard-ol.yaml"
    fi
    stdout_and_log_to_file "Kubernetes version: ${K8S_VERSION} and dashboard yaml file: ${dashboard}"
    if [ -n "${registry}" ]; then
        cat ${dashboard} | sed "s|${REGISTRY}|${registry}|g" > /tmp/dashboard.yaml
        dashboard="/tmp/dashboard.yaml"
    fi
    ${KUBECTL} create -f ${dashboard}

    if ! systemctl -q is-enabled kubectl-proxy; then
        stdout_and_log_to_file "Enabling kubectl-proxy.service ..."
        systemctl enable kubectl-proxy > /dev/null 2>&1
    fi

    if systemctl -q is-active kubectl-proxy; then
        stdout_and_log_to_file "Restarting kubectl-proxy.service ..."
        systemctl restart kubectl-proxy > /dev/null 2>&1
    else
        stdout_and_log_to_file "Starting kubectl-proxy.service ..."
        systemctl start kubectl-proxy > /dev/null 2>&1
    fi
};

: '
    Upgrade kubelet to 1.11 version
'
function upgrade_kubelet11 {
  local registry=${1:-${REGISTRY}}

  rm -rf ${KUBELET_EXTRA}
  stdout_and_log_to_file "Upgrading kubelet and kubectl now ... "
  check_v1_11

  if [ ! -f "/var/lib/kubelet/kubeadm-flags.env" ]; then
    cat <<EOF >/var/lib/kubelet/kubeadm-flags.env
KUBELET_KUBEADM_ARGS=--cgroup-driver=cgroupfs --cni-bin-dir=/opt/cni/bin --cni-conf-dir=/etc/cni/net.d --network-plugin=cni
EOF
        chmod 644 /var/lib/kubelet/kubeadm-flags.env
  fi

  registry_location11 "${registry}"
}

: '
    Checking the kubelet version if that is compatible
'
# UPGRADE PART START
function upg_pre_v1_11 {
    local kbadmv=$(rpm -qa kubeadm | awk -F- '{print $2 "-" $3}')
    local kbletv=$(rpm -qa kubelet | awk -F- '{print $2 "-" $3}')
    if [ "${kbadmv}" = "${kbletv}" ]; then
        stdout_and_log_to_file "[ERROR] Detected kubelet version ${kbletv}"
        stdout_and_log_to_file "Please downgrade kubelet to previous lower version before upgrade"
        exit 1
    fi
};

: '
    Upgrade worker node from 1.9 version to 1.12 version
'
function upgrade_worker12 {
    local registry=${1:-${REGISTRY}}

    #upg_pre_v1_11

    local pod=$(docker ps --format "{{.Names}}" | grep -v k8s_POD)
    if [ -z "${pod}" ]; then
        stdout_and_log_to_file "[ERROR] No kubernetes PODs are found in this node, nothing to upgrade"
        exit 1
    fi
    local kubesystem=$(docker ps --format "{{.Names}}" | grep -v k8s_POD | grep "kube-system")
    # if there is no kube-system it could be that other k8s pods are running so allow upgrade
    for i in $pod; do
        local found=true
        for y in ${kubesystem}; do
                if [ "${y}" = "${i}" ]; then
            found=false
            break
                fi
        done
        if [ "${found}" = "true" ]; then
                stdout_and_log_to_file "[ERROR] Upgrade detects that there is at least one user workload that is running in this node"
                stdout_and_log_to_file "        Please properly drain this node before upgrading !"
                exit 1
        fi
    done

    upgrade_kubelet12 "${registry}"

    ${KUBEADM} upgrade node config --kubelet-version $(kubelet --version | cut -d ' ' -f 2) | tee -a ${LOG_FILE}
    check_exit_code $?  "Applying kubeadm upgrade on worker node, k8s version ${K8S_VERSION}"
    systemctl restart kubelet

    mkdir -p /etc/kubernetes/manifests

    echo $(date) > /usr/local/share/kubeadm/.upgrade-worker
    stdout_and_log_to_file "[WORKER NODE UPGRADED SUCCESSFULLY]" | tee -a ${LOG_FILE}
};

: '
    Upgrade prerequisite before applying upgrade with kubeadm apply command
'
function upgrade_prereqs {
    local check=""

    if [ ! -d "/etc/kubernetes/manifests" ]; then
        stdout_and_log_to_file "[ERROR] Directory /etc/kubernetes/manifests doesn't exist"
        exit 1
    fi

    if [ ! -f "${KUBECONFIG}" ]; then
        stdout_and_log_to_file "[ERROR] KUBECONFIG ${KUBECONFIG} is not found, this file is needed for upgrade!"
        exit 1
    fi

    stdout_and_log_to_file "Checking whether api-server is using image lower than 1.9"
    check=$(grep image: /etc/kubernetes/manifests/kube-apiserver.yaml | sed 's|.*kube-apiserver.*\:||' | awk -F. '{print $2 $3}')
    if [ "${check}" = "$(echo ${K8S_VERSION} | awk -F. '{print $2 $3}')" ]; then
        stdout_and_log_to_file "[ERROR] Your clusters seem to be already in the latest version"
        exit 1
    else
        check=$(echo ${check} | awk -F- '{print $1}')
        if [ "${check}" -lt "90" ]; then
            stdout_and_log_to_file "[ERROR] Upgrade doesn't support cluster < 1.9"
            exit 1
        fi
    fi

    # fixup kubelet
    if [ "${K8S_VERSION}" = "${K8S_V10_IMAGE}" ]; then
        if [ ! -f "/etc/systemd/system/kubelet.service.d/10-kubeadm.conf" ]; then
            cp /usr/local/share/kubeadm/10-kubeadm-orig.conf /etc/systemd/system/kubelet.service.d/10-kubeadm.conf
            chmod 755 /etc/systemd/system/kubelet.service.d/10-kubeadm.conf
            systemctl daemon-reload
        fi
    else
        upg_pre_v1_11
    fi


};

: '
    Check if the system has access to the registry and has the images with specified version
'
function check_access_to_repo() {
    local registry="${1:-${REGISTRY}}"
    stdout_and_log_to_file "Checking access to ${registry} for update"
    check_repo_image="kube-proxy-amd64:${K8S_VERSION}"
    if [ "${K8S_VERSION}" = "${K8S_V12_IMAGE}" ]; then
        check_repo_image="kube-proxy:${K8S_V12_IMAGE}"
    fi

    check_repo "${registry}" "${check_repo_image}"
};
: '
    This is the main function of upgrading master node.
    K8S_VERSION is expected to be filled in before calling this function
'
function upgrade_master_main {
    local registry="${1:-${REGISTRY}}"
    local k8s_share="/usr/local/share/kubeadm"
    stdout_and_log_to_file "Upgrading pre-requisite"
    upgrade_prereqs
    stdout_and_log_to_file "Upgrading pre-requisite done"
    sleep 15 # some delay to account for kubelet restart for now
    stdout_and_log_to_file "Checking cluster health ..."
    local components="controller-manager scheduler dns apiserver etcd"
    for i in ${components}; do
        local status=$(${KUBECTL} get po | awk -v comp=${i} '$1 ~ comp {print $3}')
        if [ "${status}" != "Running" ]; then
            for((j=0; j<${TIMEOUT}; j++)); do
                sleep 1
                printf "."
                local restatus=$(${KUBECTL} get po | awk -v comp=${i} '$1 ~ comp {print $3}')
                if [ "${restatus}" = "Running" ]; then
                    break
                fi
            done
            if [ "${j}" -eq ${TIMEOUT} ]; then
                stdout_and_log_to_file "[ERROR] Components ${i} is not in the Running state"
                stdout_and_log_to_file "        Please check cluster health and perform upgrade again"
                exit 1
            fi
        fi
        sleep 1
    done
    echo ""

    ${KUBEADM} config view > /var/run/kubeadm/kubeadm-cfg
    sed -i "s|imageRepository:.*|imageRepository: ${registry}|" /var/run/kubeadm/kubeadm-cfg

    if [ "${K8S_VERSION}" = "${K8S_V12_IMAGE}" ]; then
        ${KUBEADM} upgrade apply ${K8S_VERSION} --config /var/run/kubeadm/kubeadm-cfg --allow-experimental-upgrades --feature-gates=CoreDNS=false -y --force | tee -a ${LOG_FILE}
        check_exit_code $?  "Applying kubeadm upgrade k8s version ${K8S_VERSION}"
        upgrade_kubelet12 "${registry}"
    elif [ "${K8S_VERSION}" = "${K8S_V11_IMAGE}" ]; then
        ${KUBEADM} upgrade apply ${K8S_VERSION} --config /var/run/kubeadm/kubeadm-cfg --allow-experimental-upgrades --feature-gates=CoreDNS=false -y --force | tee -a ${LOG_FILE}
        check_exit_code $?  "Applying kubeadm upgrade k8s version ${K8S_VERSION}"
        upgrade_kubelet11 "${registry}"
    elif [ "${K8S_VERSION}" = "${K8S_V10_IMAGE}" ]; then
        ${KUBEADM} upgrade apply ${K8S_VERSION} --config /var/run/kubeadm/kubeadm-cfg --allow-experimental-upgrades -y --force | tee -a ${LOG_FILE}
        check_exit_code $?  "Applying kubeadm upgrade k8s version ${K8S_VERSION}"
        registry_location10 "${registry}"
    fi

    # in one master upgrade dashboard could be terminated so don't need to reinstall
    local dashboard=$(${KUBECTL} get pod | awk '/kubernetes-dashboard/ && $3 != "Terminating" {print $1}')

    if [ -z "${dashboard}" ]; then
        stdout_and_log_to_file "[INSTALLING DASHBOARD NOW]"
        install_dashboard "${registry}"
    fi

    echo "Upgrading kubeadm(${K8S_VERSION}) $(date)" > ${k8s_share}/.upgrade-cluster
};

: '
    Create backup of the cluster
'
function backup_cluster() {
    if ! is_master; then
        echo "Backup must be done on master node. exiting."
        exit 1;
    fi

    if $(${KUBECTL} get cs>/dev/null 2>&1); then
	      echo "[ERROR] Please temporarily stop the cluster for backup/restore operation"
	      echo "        by doing the following:"
	      echo "        # kubeadm-setup.sh stop"
	      exit 1
    fi

    echo "Backing up cluster"
    local BACKUP_FOLDER=${1:-}
    if kubeadm-setup.sh backup ${BACKUP_FOLDER}; then
        BACKUP_FILE=$(ls -lrt ${BACKUP_FOLDER}/master-backup-v1.9.* | tail -n1 | grep -Eow master.*)
        echo "Storing meta-data to backup file $BACKUP_FILE"
        rpm -q kubelet kubectl kubeadm  > .version-info
        tar rvfP $BACKUP_FOLDER/$BACKUP_FILE .version-info
        rm .version-info
        echo "Backup creation successful :)"
    else
        echo "Backup failed!"
	exit 1
    fi
};

: '
    Restore the cluster from the previous backed up  tar file.
'
function restore_cluster() {
    if ! is_master; then
        echo "Restore must be done on master node. exiting."
        exit 1;
    fi

    if $(${KUBECTL} get cs>/dev/null 2>&1); then
	      echo "[ERROR] Please temporarily stop the cluster for backup/restore operation"
	      echo "        by doing the following:"
	      echo "        # kubeadm-setup.sh stop"
	      exit 1
    fi

    echo "Restoring the cluster"
    local RESTORE_FOLDER=${1:-}
    if [ ! -f "${RESTORE_FOLDER}" ]; then
        echo "[ERROR] tar file of ${RESTORE_FOLDER} doesn't exist ..."
        exit 1
    fi
    local KUBE_VERSIONS=$(tar -axf ${RESTORE_FOLDER} .version-info -O 2>&1)

    if [[ -z "${KUBE_VERSIONS}" ]]; then
        echo "Restore failed! Error reading ${RESTORE_FOLDER} and downgrading kubecomponents"
        exit 1
    fi

    if yum downgrade -y ${KUBE_VERSIONS}; then
        if kubeadm-setup.sh restore ${RESTORE_FOLDER}; then
            echo "Restore successful :)"
        else
            echo "Restore failed!"
            exit 1
        fi
    else
        echo "Downgrade failed! Error downgrading the packages: ${KUBE_VERSIONS}"
        exit 1
    fi
};

: '
    Upgrade worker node from 1.9 to 1.12 version
'
function upgrade_worker() {
    stdout_and_log_to_file "Upgrading worker node"
    stdout_and_log_to_file "Updating kubeadm package"
    local registry="${KUBE_REPO_PREFIX:-${REGISTRY}}"
    K8S_VERSION=${K8S_V12_IMAGE}
    check_access_to_repo ${registry}
    check_package_in_yum kubeadm-${K8S_V12_VERSION}
    stdout_and_log_to_file "Upgrading kubeadm to latest version"
    timeout ${TIMEOUT} yum update -y kubeadm-${K8S_V12_VERSION} | tee -a ${LOG_FILE}
    upgrade_worker12 ${registry}
};
: '
    Upgrading the master node from 1.9 to 1.12 version through intermediate 1.10 and 1.11 version
'
function upgrade_master() {
    local registry="${KUBE_REPO_PREFIX:-${REGISTRY}}"
    stdout_and_log_to_file "Upgrading master node"
    check_package_in_yum kubeadm-${K8S_V10_VERSION}
    check_package_in_yum kubeadm-${K8S_V11_VERSION}
    check_package_in_yum kubeadm-${K8S_V12_VERSION}
    K8S_VERSION=${K8S_V10_IMAGE}
    check_access_to_repo ${registry}
    K8S_VERSION=${K8S_V11_IMAGE}
    check_access_to_repo ${registry}
    K8S_VERSION=${K8S_V12_IMAGE}
    check_access_to_repo ${registry}
    K8S_VERSION=${K8S_V10_IMAGE}
    stdout_and_log_to_file "Upgrading kubeadm,kubectl and kubelet to ${K8S_V10_VERSION} version"
    timeout ${TIMEOUT} yum install -y kubeadm-${K8S_V10_VERSION} kubectl-${K8S_V10_VERSION} kubelet-${K8S_V10_VERSION} | tee -a ${LOG_FILE}
    check_exit_code $? "Updating kube* to 1.10.5"
    upgrade_master_main ${registry}
    stdout_and_log_to_file "Upgrading kubeadm to 1.11.3 version"
    timeout ${TIMEOUT} yum install -y kubeadm-${K8S_V11_VERSION} | tee -a ${LOG_FILE}
    check_exit_code $? "Updating kubeadm-1.11.3"
    K8S_VERSION=${K8S_V11_IMAGE}
    upgrade_master_main ${registry}
    stdout_and_log_to_file "Upgrading kubelet and kubectl to 1.11.3 version"
    timeout ${TIMEOUT} yum install -y kubelet-${K8S_V11_VERSION} kubectl-${K8S_V11_VERSION} | tee -a ${LOG_FILE}
    check_exit_code $? "Updating kubelet/kubectl to 1.11.3"
    stdout_and_log_to_file "Upgrading kubeadm to ${K8S_V12_VERSION} version"
    timeout ${TIMEOUT} yum install -y kubeadm-${K8S_V12_VERSION}
    check_exit_code $? "Updating kubeadm"
    K8S_VERSION=${K8S_V12_IMAGE}
    upgrade_master_main ${registry}
    stdout_and_log_to_file "[INFO] Upgrading master node done successfully"
    stdout_and_log_to_file "[INFO] Flannel is not upgraded yet. Please run 'kubeadm-upgrade.sh upgrade --flannel' to upgrade flannel"
    stdout_and_log_to_file "[INFO] Dashboard is not upgraded yet. Please run 'kubeadm-upgrade.sh upgrade --dashboard' to upgrade dashboard"
};

: '
    Print a progress bar.
'
function progress_bar() {
    spin='-\|/'
    i=0
    while kill -0 $1 2>/dev/null
    do
        i=$(( (i+1) %4 ))
        printf "\r${spin:$i:1}"
        sleep .2
    done
};

: '
    Check if the script is run as a root user.
'
function check_user() {
    if [[ $EUID -ne 0 ]]; then
       do_exit "This script must be run as root, exiting" 1
    fi
};

: '
'
function check_k8s_version() {
    package=$1
    target=$2
    if ! isinstalled $package; then
        do_exit "${package} not installed. Exiting the script" 1
    fi
    current_version=$(get_package_version ${package})
    echo "$package version $current_version"
    if [[ "${current_version}" != "${target}" ]]; then
       do_exit "${package} version must be ${target}, exiting" 1
    fi
};

: '
'
function check_system() {
    no_cpu=$(cat /proc/cpuinfo | awk '/^processor/{print $3}' | wc -l)
    stdout_and_log_to_file "Number of cpu present in this system ${no_cpu}"
    if [[ "${no_cpu}" -lt 2 ]]; then
        do_exit "Not enough CPU resource. Recommended # of CPUs 2 or more. Exiting" 1
    fi
    memt=$(cat /proc/meminfo| grep MemTotal: | awk -F ' ' '{print $2}')
    memMB=$((memt/1024))
    stdout_and_log_to_file "Total memory on this system: ${memMB}MB"
    if [[ $memMB -lt 2048 ]]; then
            stdout_and_log_to_file "This system has less memory resources. Upgrade might fail or cluster may not be healthy. Recommended memory 2GB or more"
    fi
    check_available_space /var/lib/docker 5
    check_available_space /var/lib/kubelet 5
    check_k8s_version "kubeadm" 1.9
    check_k8s_version "kubectl" 1.9
    check_k8s_version "kubelet" 1.9
    dev_repo=ol7_developer
    addon_repo=ol7_addons
    if repo_enabled ${addon_repo}; then
        stdout_and_log_to_file "${addon_repo} repo enabled"
    else
        stdout_and_log_to_file "${addon_repo} repo disabled. Will try to enable the repo if present"
        yum-config-manager --enable ${addon_repo} | tee -a ${LOG_FILE}
        check_exit_code $? "Enabling ${addon_repo} repo"
    fi

};

: '
    Check whether the host is master or worker. For master this function returns true otherwise false
'
function is_master() {
    MASTER_API_KEY_FILE=/etc/kubernetes/pki/apiserver.key
    if [ ! -f ${MASTER_API_KEY_FILE} ]; then
        false
    else
        true
    fi
};

: '
    Update the yum cache. This function should be called up updating the yum repo file.
'
function update_yum_cache() {
    stdout_and_log_to_file "Updating yum cache"
    yum makecache fast >/dev/null &
    pid=$!
    progress_bar $pid
    stdout_and_log_to_file "Updating the yum cache done"
};

function confirm_upgrade() {
    echo "[WARNING] This action will upgrade this node to latest version"
    echo "[WARNING] The cluster will be upgraded through intermediate versions which are unsupported"
    echo "[WARNING] You must take backup before upgrading the cluster as upgrade may fail"
    echo "  Please select 1 (continue) or 2 (abort) :"
    select choice in "continue" "abort"; do
        case ${choice} in
            "continue" )
            break
            ;;
            "abort" )
            exit
            ;;
        esac
    done
};

TIMEOUT=500
REGISTRY="container-registry.oracle.com/kubernetes"
KUBECONFIG="/etc/kubernetes/admin.conf"
KUBELET_EXTRA="/etc/systemd/system/kubelet.service.d/20-pod-infra-image.conf"
KUBELET_EXTRA_NEW="/etc/sysconfig/kubelet"
K8S_VERSION="v1.9.1"
NOPROXY="env -u http_proxy -u https_proxy -u HTTP_PROXY -u HTTPS_PROXY"
KUBECTL="${NOPROXY} kubectl --kubeconfig=${KUBECONFIG} -n kube-system"
KUBEADM="${NOPROXY} kubeadm"
K8S_V12_VERSION=1.12.7-1.1.2.el7
K8S_V12_IMAGE="v1.12.7"
K8S_V10_VERSION=1.10.5-2.5.4.el7
K8S_V10_IMAGE="v1.10.5-2"
K8S_V11_VERSION=1.11.3-2.5.2.el7
K8S_V11_IMAGE="v1.11.3-2"
check_user
LOG_FOLDER=/var/log/kubeadm-upgrade
today=$(date "+%Y-%m-%d-%H:%M:%S")
echo "-- Running upgrade script---"
mkdir -p ${LOG_FOLDER}
LOG_FILE=${LOG_FOLDER}/k8s-upgrade-${today}.log
export FORCE_UPDATE_12_FROM_9=1
# Parse the commands
case "${1:-}" in
    upgrade)
    shift

    if [ ! -z "${1}" ]; then
	upgrade_addons "${1}"
	exit 0
    fi

    check_system
    confirm_upgrade
    if is_master; then
        upgrade_master
    else
        upgrade_worker
    fi
    ;;
    backup)
    shift
    backup_cluster ${1:-}
    ;;
    restore)
    shift
    restore_cluster ${1:-}

    ;;
    *)
    print_usage
    ;;
esac
